/*******************************************************************************

  Example Plugin Library (DLL) for DIMIN Viewer n5 (DIV5)

  Programmer: Dima V. Fedorov Levit <dimin@dimin.net> <http://www.dimin.net/>

  History:
    12/05/2000 20:36:48 - First Creation
    03/25/2001 17:11:55 - Conversion to v2.1
    06/20/2004 01:10:00 - DIV plugin interface v5
    05/26/2005 16:09:00 - more filters

  ver: 2

*******************************************************************************/

#include <windows.h>

#include "div5_xtdFilters.h"

TDIVPluginHeader PluginHeader = {
  sizeof( TDIVPluginHeader ),
  5,   // plugin interface version - MUST BE 5
  DIV_PLUGIN_TYPE_FILTER,
  3,   // plugin version
  13,  // 10 filters
  "DIMIN Extended Filters"
};

//------------------------------------------------------------------------------
// About procedure
//------------------------------------------------------------------------------
void __stdcall aboutPlugin( HWND hwndParent, TDIVParameters *pars ) 
{
  MessageBox(GetFocus(), "DIV5 Extended Filters v3", "About Plug-in", MB_OK | MB_ICONEXCLAMATION);
}

TDIVFilterHeader headerDevider = {
  sizeof(TDIVFilterHeader),
  "-",
  "-",
  NULL, // filter
  0,       // no bpp needed
  0,    // arguments, if NULL equals to usePreview = 0
  NULL
};


//------------------------------------------------------------------------------
// Filter 1 rotates left provided image by 90 degrees
//------------------------------------------------------------------------------

TDIVFilterHeader FilterHeader1 = {
  sizeof(TDIVFilterHeader),
  "DIV_IMAGE_ROTATE_90_LEFT",
  "Rotate by 90 degrees to the Left",
  filterRL, // filter
  777,       // no bpp needed
  NULL,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

//------------------------------------------------------------------------------
// Filter 2 rotates right provided image by 90 degrees
//------------------------------------------------------------------------------

TDIVFilterHeader FilterHeader2 = {
  sizeof(TDIVFilterHeader),
  "DIV_IMAGE_ROTATE_90_RIGHT",
  "Rotate by 90 degrees to the Right",
  filterRR, // filter
  777,       // no bpp needed
  NULL,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

//------------------------------------------------------------------------------
// Filter 3 find edges
//------------------------------------------------------------------------------

TDIVArguments filter3Args = {
  1,          // 0 - no preview window, 1 - DIV will open preview
  0,          // number of arguments in the array
  NULL, // pointer to array of argumants
  NULL, // custom controls creation INIT
  NULL, // custom controls creation QUIT
};

TDIVFilterHeader FilterHeader3 = {
  sizeof(TDIVFilterHeader),
  "DIV_FILTER_EDGES_FIND",
  "Find Edges",
  filterFindEdg, // filter
  666,       // no bpp needed
  &filter3Args,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};


//------------------------------------------------------------------------------
// Filter 4 median
//------------------------------------------------------------------------------

/*
TDIVArguments filter01Args = {
  1,          // 0 - no preview window, 1 - DIV will open preview
  1,          // number of arguments in the array
  &amountFilterArgsArr, // pointer to array of argumants
  NULL, // custom controls creation INIT
  NULL, // custom controls creation QUIT
};

  */

TDIVTagedArgument amountFilterArgsArr(
  1000,         // id number, plugin may use ids above 1000 as private
  DIM_TAG_LONG, // TDimTagType
  1,            // number of records of type tagType pointed by tagData
  "Amount:",    // text description of the argument
  TDIVDataValue(1),   // pointer to data, if data size is <= 4 bytes
                // then it's directly stored into pointer
  TDIVDataValue(1),   // pointer to min val, used if needed to define
                // a special value retrieval, e.g. -100 .. 100
  TDIVDataValue(50)   // pointer to max val
);

TDIVArguments filter4Args = {
  1,          // 0 - no preview window, 1 - DIV will open preview
  1,          // number of arguments in the array
  &amountFilterArgsArr, // pointer to array of argumants
  NULL, // custom controls creation INIT
  NULL, // custom controls creation QUIT
};

TDIVFilterHeader FilterHeader4 = {
  sizeof(TDIVFilterHeader),
  "DIV_FILTER_MEDIAN",
  "Median",
  filterMedian, // filter
  666,       // no bpp needed
  &filter4Args,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

TDIVFilterHeader FilterHeaderMedStrong = {
  sizeof(TDIVFilterHeader),
  "DIV_FILTER_MEDIAN_STRONG",
  "Median Strong",
  filterMedianStrong, // filter
  666,       // no bpp needed
  &filter4Args,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

//------------------------------------------------------------------------------
// Filter 5-6 blur 3x3, Gaussian
//------------------------------------------------------------------------------

TDIVTagedArgument blurFilterArgsArr (
  1000,         // id number, plugin may use ids above 1000 as private
  DIM_TAG_LONG, // TDimTagType
  1,            // number of records of type tagType pointed by tagData
  "Amount:",    // text description of the argument
  1,   // pointer to data, if data size is <= 4 bytes
                // then it's directly stored into pointer
  1,   // pointer to min val, used if needed to define
                // a special value retrieval, e.g. -100 .. 100
  10  // pointer to max val
);

TDIVArguments blurFiltersArgs = {
  1,          // 0 - no preview window, 1 - DIV will open preview
  1,          // number of arguments in the array
  &blurFilterArgsArr, // pointer to array of argumants
  NULL, // custom controls creation INIT
  NULL, // custom controls creation QUIT
};

TDIVFilterHeader FilterHeader5 = {
  sizeof(TDIVFilterHeader),
  "DIV_FILTER_BLUR_SMOOTH",
  "Smooth",
  filterBlur3x3, // filter
  666,       // no bpp needed
  &blurFiltersArgs,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

TDIVFilterHeader FilterHeader6 = {
  sizeof(TDIVFilterHeader),
  "DIV_FILTER_BLUR_GAUSSIAN",
  "Gaussian Blur",
  filterBlurGauss, // filter
  666,       // no bpp needed
  &blurFiltersArgs,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};


//------------------------------------------------------------------------------
// Filter 7-8 RGB<->HSL
//------------------------------------------------------------------------------

TDIVFilterHeader FilterHeader7 = {
  sizeof(TDIVFilterHeader),
  "DIV_COLOR_SPACE_RGB_HSL",
  "Convert RGB->HSL (normalized 0-255)",
  filterRGBtoHSL, // filter
  24,       // no bpp needed
  NULL,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

TDIVFilterHeader FilterHeader8 = {
  sizeof(TDIVFilterHeader),
  "DIV_COLOR_SPACE_HSL_RGB",
  "Convert HSL->RGB (normalized 0-255)",
  filterHSLtoRGB, // filter
  24,       // no bpp needed
  NULL,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

//------------------------------------------------------------------------------
// Filter 9-10 RGB<->YUV
//------------------------------------------------------------------------------

TDIVFilterHeader FilterHeader9 = {
  sizeof(TDIVFilterHeader),
  "DIV_COLOR_SPACE_RGB_YUV",
  "Convert RGB->YUV (normalized 0-255)",
  filterRGBtoYUV, // filter
  24,       // no bpp needed
  NULL,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

TDIVFilterHeader FilterHeader10 = {
  sizeof(TDIVFilterHeader),
  "DIV_COLOR_SPACE_YUV_RGB",
  "Convert YUV->RGB (normalized 0-255)",
  filterYUVtoRGB, // filter
  24,       // no bpp needed
  NULL,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};


//------------------------------------------------------------------------------
// Filter 11-12 RGB<->LAB
//------------------------------------------------------------------------------

TDIVFilterHeader FilterHeader11 = {
  sizeof(TDIVFilterHeader),
  "DIV_COLOR_SPACE_RGB_LAB",
  "Convert RGB->LAB (normalized 0-255)",
  filterRGBtoLAB, // filter
  24,       // no bpp needed
  NULL,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

TDIVFilterHeader FilterHeader12 = {
  sizeof(TDIVFilterHeader),
  "DIV_COLOR_SPACE_LAB_RGB",
  "Convert LAB->RGB (normalized 0-255)",
  filterLABtoRGB, // filter
  24,       // no bpp needed
  NULL,    // arguments, if NULL equals to usePreview = 0
  aboutPlugin
};

//***************************************************************************
// DLL entrie
//***************************************************************************

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
  switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}


TDIVPluginHeader* __stdcall divGetPluginHeader()
{
   return &PluginHeader;
}

TDIVFilterHeader* __stdcall divGetFilterHeader(int I)
{
  switch (I)
  {
    case 0: return &FilterHeader1;
    case 1: return &FilterHeader2;

    case 2: return &headerDevider;

    case 3: return &FilterHeader3;
    case 4: return &FilterHeader4;
    case 5: return &FilterHeaderMedStrong;

    case 6: return &FilterHeader5;
    case 7: return &FilterHeader6;

    case 8: return &headerDevider;

    case 9: return &FilterHeader7;
    case 10: return &FilterHeader8;
    case 11: return &FilterHeader9;
    case 12: return &FilterHeader10;
    //case 10: return &FilterHeader11;
    //case 11: return &FilterHeader12;
    default : return NULL;
  }
}

TDivPluginFilter __stdcall divGetFilter(int I)
{
  switch (I)
  {
    case 0: return filterRL;
    case 1: return filterRR;

    case 2: return NULL;

    case 3: return filterFindEdg;
    case 4: return filterMedian;
    case 5: return filterMedianStrong;
    case 6: return filterBlur3x3;
    case 7: return filterBlurGauss;

    case 8: return NULL;

    case 9: return filterRGBtoHSL;
    case 10: return filterHSLtoRGB;
    case 11: return filterRGBtoYUV;
    case 12: return filterYUVtoRGB;
    //case 13: return filterRGBtoLAB;
    //case 14: return filterLABtoRGB;
    default: return NULL;
  }
}

//---------------------------------------------------------------------------
