/*******************************************************************************

  Convolution for DIMIN Viewer n5 (DIV5)

  Programmer: Dima V. Fedorov Levit <dimin@dimin.net> <http://www.dimin.net/>

  History:
    05/25/2005 20:37 - First Creation
  
  ver: 1  
*******************************************************************************/

#include <stdio.h>
#include <windows.h>

#include "DIVPluginDefs.h"
#include "DIVFilterDefs.h"
#include "DIVFilterUtils.h"


//******************************************************************************
// Convolution of 3x3 kernel
//******************************************************************************

void applyKernel33( TDIVParameters *par, BITMAP *inbmp, int **k, int bias, int pbias, char *txt )
{
  unsigned char *p0, *p1, *p2;
  uint x, y;  
  unsigned int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;


  uint pB=0, pE=w-1;
  uint pI=0;

  if (bpp == 8)
  {

    // top
    y = 0;
    p0 = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );
    p2 = (unsigned char *) getHBScanLine( h-y-2, inp, w, bpp );

    findLineBegEndTopBot( par, w, y, pI, pB, pE);
    for (x=pB+1; x<pE; x++)
      p0[x] = ( ( k[2][1]*p0[x-1] ) + ( k[2][2]*p0[x] ) + ( k[2][3]*p0[x+1] ) +
                ( k[3][1]*p2[x-1] ) + ( k[3][2]*p2[x] ) + ( k[3][3]*p2[x+1] ) ) / pbias;

    for (y=1; y<h-1; y++)
    {
      showProgress( par, y, h, txt );
      if ( testAbort( par ) == 1) break;
      findLineBegEndTopBot( par, w, y, pI, pB, pE );

      p1 = (unsigned char *) getHBScanLine( h-y-0, inp, w, bpp );      
      p0 = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );
      p2 = (unsigned char *) getHBScanLine( h-y-2, inp, w, bpp );

      // left
      x = 0;
      if (pB >= x)
      p0[x] = ( ( k[1][2]*p1[x] ) + ( k[1][3]*p1[x+1] ) +
                ( k[2][2]*p0[x] ) + ( k[2][3]*p0[x+1] ) +
                ( k[3][2]*p2[x] ) + ( k[3][3]*p2[x+1] ) ) / pbias;

      // center
      for (x=pB+1; x<pE; x++)
        p0[x] = ( ( k[1][1]*p1[x-1] ) + ( k[1][2]*p1[x] ) + ( k[1][3]*p1[x+1] ) +
                  ( k[2][1]*p0[x-1] ) + ( k[2][2]*p0[x] ) + ( k[2][3]*p0[x+1] ) +
                  ( k[3][1]*p2[x-1] ) + ( k[3][2]*p2[x] ) + ( k[3][3]*p2[x+1] ) ) / bias;

      // right
      x = w-1;
      if (pE <= x)
      p0[x] = ( ( k[1][1]*p1[x-1] ) + ( k[1][2]*p1[x] ) + 
                ( k[2][1]*p0[x-1] ) + ( k[2][2]*p0[x] ) + 
                ( k[3][1]*p2[x-1] ) + ( k[3][2]*p2[x] ) ) / pbias;

    } 

    // bottom

    y  = h-1;
    p1 = (unsigned char *) getHBScanLine( h-y-0, inp, w, bpp );      
    p0 = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );
    findLineBegEndTopBot( par, w, y, pI, pB, pE);
    for (x=pB+1; x<pE; x++)
      p0[x] = ( ( k[1][1]*p1[x-1] ) + ( k[1][2]*p1[x] ) + ( k[1][3]*p1[x+1] ) +
                ( k[2][1]*p0[x-1] ) + ( k[2][2]*p0[x] ) + ( k[2][3]*p0[x+1] ) ) / pbias;

  } // 8 bpp


  if (bpp == 24)
  {

    // top 
    y = 0;
    p0 = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );
    p2 = (unsigned char *) getHBScanLine( h-y-2, inp, w, bpp );

    findLineBegEndTopBot( par, w, y, pI, pB, pE);
    for (x=(pB+1)*3; x<pE*3; x++)
      p0[x] = ( ( k[2][1]*p0[x-3] ) + ( k[2][2]*p0[x] ) + ( k[2][3]*p0[x+3] ) +
                ( k[3][1]*p2[x-3] ) + ( k[3][2]*p2[x] ) + ( k[3][3]*p2[x+3] ) ) / pbias;

    for (y=1; y<h-1; y++)
    {
      showProgress( par, y, h, txt );
      if ( testAbort( par ) == 1) break;
      findLineBegEndTopBot( par, w, y, pI, pB, pE );

      p1 = (unsigned char *) getHBScanLine( h-y-0, inp, w, bpp );      
      p0 = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );
      p2 = (unsigned char *) getHBScanLine( h-y-2, inp, w, bpp );

      // left
      x = 0;
      if (pB >= x)
      for (x=0; x<3; x++)
      p0[x] = ( ( k[1][2]*p1[x] ) + ( k[1][3]*p1[x+3] ) +
                ( k[2][2]*p0[x] ) + ( k[2][3]*p0[x+3] ) +
                ( k[3][2]*p2[x] ) + ( k[3][3]*p2[x+3] ) ) / pbias;

      // center
      for (x=(pB+1)*3; x<pE*3; x++)
        p0[x] = ( ( k[1][1]*p1[x-3] ) + ( k[1][2]*p1[x] ) + ( k[1][3]*p1[x+3] ) +
                  ( k[2][1]*p0[x-3] ) + ( k[2][2]*p0[x] ) + ( k[2][3]*p0[x+3] ) +
                  ( k[3][1]*p2[x-3] ) + ( k[3][2]*p2[x] ) + ( k[3][3]*p2[x+3] ) ) / bias;

      // right
      x = w-1;
      if (pE <= x)
      for (x=w*3-3; x<w*3; x++)
      p0[x] = ( ( k[1][1]*p1[x-3] ) + ( k[1][2]*p1[x] ) + 
                ( k[2][1]*p0[x-3] ) + ( k[2][2]*p0[x] ) + 
                ( k[3][1]*p2[x-3] ) + ( k[3][2]*p2[x] ) ) / pbias;
    } 

    // bottom
    y  = h-1;
    p1 = (unsigned char *) getHBScanLine( h-y-0, inp, w, bpp );      
    p0 = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );
    findLineBegEndTopBot( par, w, y, pI, pB, pE);
    for (x=(pB+1)*3; x<pE*3; x++)
      p0[x] = ( ( k[1][1]*p1[x-3] ) + ( k[1][2]*p1[x] ) + ( k[1][3]*p1[x+3] ) +
                ( k[2][1]*p0[x-3] ) + ( k[2][2]*p0[x] ) + ( k[2][3]*p0[x+3] ) ) / pbias;

  } // 24 bpp


}

void applyKernel55( TDIVParameters *par, BITMAP *inbmp, int **k, int bias, int bias1, int bias2, char *txt )
{
  unsigned char *p0, *p1, *p2, *p3, *p4;
  int x, y;  
  int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;

  int pB=0, pE=w-1;
  uint pI=0;

  if (bpp == 8)
  {

    // TOP
    y = 0;
    findLineBegEndTopBotI( par, w, y, pI, pB, pE );
    p0 = (unsigned char *) getHBScanLine( h-1-y-0, inp, w, bpp );
    p2 = (unsigned char *) getHBScanLine( h-1-y-1, inp, w, bpp );
    p4 = (unsigned char *) getHBScanLine( h-1-y-2, inp, w, bpp );

    if (pE-1 >= pB+2)    
    for (x=pB+2; x<pE-1; x++)
      p0[x] = 
      ( 
        ( k[3][1]*p0[x-2] ) + ( k[3][2]*p0[x-1] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+1] ) + ( k[3][5]*p0[x+2] ) +
        ( k[4][1]*p2[x-2] ) + ( k[4][2]*p2[x-1] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+1] ) + ( k[4][5]*p2[x+2] ) +
        ( k[5][1]*p4[x-2] ) + ( k[5][2]*p4[x-1] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+1] ) + ( k[5][5]*p4[x+2] )
      ) / bias2;

    y=1;
    findLineBegEndTopBotI( par, w, y, pI, pB, pE );
    p1 = (unsigned char *) getHBScanLine( h-1-y+1, inp, w, bpp );      
    p0 = (unsigned char *) getHBScanLine( h-1-y-0, inp, w, bpp );
    p2 = (unsigned char *) getHBScanLine( h-1-y-1, inp, w, bpp );
    p4 = (unsigned char *) getHBScanLine( h-1-y-2, inp, w, bpp );

    if (pE-1 >= pB+2)        
    for (x=pB+2; x<pE-1; x++)
      p0[x] = 
      ( 
        ( k[2][1]*p1[x-2] ) + ( k[2][2]*p1[x-1] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+1] ) + ( k[2][5]*p1[x+2] ) +
        ( k[3][1]*p0[x-2] ) + ( k[3][2]*p0[x-1] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+1] ) + ( k[3][5]*p0[x+2] ) +
        ( k[4][1]*p2[x-2] ) + ( k[4][2]*p2[x-1] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+1] ) + ( k[4][5]*p2[x+2] ) +
        ( k[5][1]*p4[x-2] ) + ( k[5][2]*p4[x-1] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+1] ) + ( k[5][5]*p4[x+2] )
      ) / bias1;

    // CENTER X
    for (y=2; y<h-2; y++)
    {
      showProgress( par, y, h, txt );
      if ( testAbort( par ) == 1) break;
      findLineBegEndTopBotI( par, w, y, pI, pB, pE );

      p3 = (unsigned char *) getHBScanLine( h-1-y-2, inp, w, bpp );      
      p1 = (unsigned char *) getHBScanLine( h-1-y-1, inp, w, bpp );      
      p0 = (unsigned char *) getHBScanLine( h-1-y-0, inp, w, bpp );
      p2 = (unsigned char *) getHBScanLine( h-1-y+1, inp, w, bpp );
      p4 = (unsigned char *) getHBScanLine( h-1-y+2, inp, w, bpp );

      // LEFT
      x = 0;
      if (pB <= x)
      p0[x] = 
      ( 
        ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+1] ) + ( k[1][5]*p3[x+2] ) +
        ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+1] ) + ( k[2][5]*p1[x+2] ) +
        ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+1] ) + ( k[3][5]*p0[x+2] ) +
        ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+1] ) + ( k[4][5]*p2[x+2] ) +
        ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+1] ) + ( k[5][5]*p4[x+2] )
      ) / bias2;

      x = 1;
      if (pB <= x)
      p0[x] = 
      ( 
        ( k[1][2]*p3[x-1] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+1] ) + ( k[1][5]*p3[x+2] ) +
        ( k[2][2]*p1[x-1] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+1] ) + ( k[2][5]*p1[x+2] ) +
        ( k[3][2]*p0[x-1] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+1] ) + ( k[3][5]*p0[x+2] ) +
        ( k[4][2]*p2[x-1] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+1] ) + ( k[4][5]*p2[x+2] ) +
        ( k[5][2]*p4[x-1] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+1] ) + ( k[5][5]*p4[x+2] )
      ) / bias1;

      // CENTER
      for (x=pB+2; x<pE-1; x++)
        p0[x] = 
        ( 
          ( k[1][1]*p3[x-2] ) + ( k[1][2]*p3[x-1] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+1] ) + ( k[1][5]*p3[x+2] ) +
          ( k[2][1]*p1[x-2] ) + ( k[2][2]*p1[x-1] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+1] ) + ( k[2][5]*p1[x+2] ) +
          ( k[3][1]*p0[x-2] ) + ( k[3][2]*p0[x-1] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+1] ) + ( k[3][5]*p0[x+2] ) +
          ( k[4][1]*p2[x-2] ) + ( k[4][2]*p2[x-1] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+1] ) + ( k[4][5]*p2[x+2] ) +
          ( k[5][1]*p4[x-2] ) + ( k[5][2]*p4[x-1] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+1] ) + ( k[5][5]*p4[x+2] )
        ) / bias;
    
      // RIGHT
      x = w-2;
      if (pE >= x)
      p0[x] = 
      ( 
        ( k[1][1]*p3[x-2] ) + ( k[1][2]*p3[x-1] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+1] ) + 
        ( k[2][1]*p1[x-2] ) + ( k[2][2]*p1[x-1] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+1] ) + 
        ( k[3][1]*p0[x-2] ) + ( k[3][2]*p0[x-1] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+1] ) + 
        ( k[4][1]*p2[x-2] ) + ( k[4][2]*p2[x-1] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+1] ) + 
        ( k[5][1]*p4[x-2] ) + ( k[5][2]*p4[x-1] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+1] )  
      ) / bias1;

      x = w-1;
      if (pE >= x)
      p0[x] = 
      ( 
        ( k[1][1]*p3[x-2] ) + ( k[1][2]*p3[x-1] ) + ( k[1][3]*p3[x] ) + 
        ( k[2][1]*p1[x-2] ) + ( k[2][2]*p1[x-1] ) + ( k[2][3]*p1[x] ) + 
        ( k[3][1]*p0[x-2] ) + ( k[3][2]*p0[x-1] ) + ( k[3][3]*p0[x] ) + 
        ( k[4][1]*p2[x-2] ) + ( k[4][2]*p2[x-1] ) + ( k[4][3]*p2[x] ) + 
        ( k[5][1]*p4[x-2] ) + ( k[5][2]*p4[x-1] ) + ( k[5][3]*p4[x] )  
      ) / bias2;

    } // for (y=2; y<h-2; y++)


    // BOTTOM
    y = h-2;
    findLineBegEndTopBotI( par, w, y, pI, pB, pE );

    p3 = (unsigned char *) getHBScanLine( h-1-y+2, inp, w, bpp );      
    p1 = (unsigned char *) getHBScanLine( h-1-y+1, inp, w, bpp );      
    p0 = (unsigned char *) getHBScanLine( h-1-y-0, inp, w, bpp );
    p2 = (unsigned char *) getHBScanLine( h-1-y-1, inp, w, bpp );

    //for (x=2; x<img->width()-2; x++)
    for (x=pB+2; x<pE-1; x++)
      p0[x] = 
      ( 
        ( k[1][1]*p3[x-2] ) + ( k[1][2]*p3[x-1] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+1] ) + ( k[1][5]*p3[x+2] ) +
        ( k[2][1]*p1[x-2] ) + ( k[2][2]*p1[x-1] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+1] ) + ( k[2][5]*p1[x+2] ) +
        ( k[3][1]*p0[x-2] ) + ( k[3][2]*p0[x-1] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+1] ) + ( k[3][5]*p0[x+2] ) +
        ( k[4][1]*p2[x-2] ) + ( k[4][2]*p2[x-1] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+1] ) + ( k[4][5]*p2[x+2] ) 
      ) / bias1;

    y = h-1;
    findLineBegEndTopBotI( par, w, y, pI, pB, pE );

    p3 = (unsigned char *) getHBScanLine( h-1-y+2, inp, w, bpp );      
    p1 = (unsigned char *) getHBScanLine( h-1-y+1, inp, w, bpp );      
    p0 = (unsigned char *) getHBScanLine( h-1-y+0, inp, w, bpp );

    for (x=pB+2; x<pE-1; x++)
      p0[x] = 
      ( 
        ( k[1][1]*p3[x-2] ) + ( k[1][2]*p3[x-1] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+1] ) + ( k[1][5]*p3[x+2] ) +
        ( k[2][1]*p1[x-2] ) + ( k[2][2]*p1[x-1] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+1] ) + ( k[2][5]*p1[x+2] ) +
        ( k[3][1]*p0[x-2] ) + ( k[3][2]*p0[x-1] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+1] ) + ( k[3][5]*p0[x+2] ) 
      ) / bias2;

  } // 8 bpp

  if (bpp == 24)
  {

    // TOP
    y = 0;
    findLineBegEndTopBotI( par, w, y, pI, pB, pE );
    p0 = (unsigned char *) getHBScanLine( h-1-y-0, inp, w, bpp );
    p2 = (unsigned char *) getHBScanLine( h-1-y-1, inp, w, bpp );
    p4 = (unsigned char *) getHBScanLine( h-1-y-2, inp, w, bpp );

    for (x=(pB+2)*3; x<(pE-1)*3; x++)   
      p0[x] = 
      ( 
        ( k[3][1]*p0[x-6] ) + ( k[3][2]*p0[x-3] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+3] ) + ( k[3][5]*p0[x+6] ) +
        ( k[4][1]*p2[x-6] ) + ( k[4][2]*p2[x-3] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+3] ) + ( k[4][5]*p2[x+6] ) +
        ( k[5][1]*p4[x-6] ) + ( k[5][2]*p4[x-3] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+3] ) + ( k[5][5]*p4[x+6] )
      ) / bias2;

    y=1;
    findLineBegEndTopBotI( par, w, y, pI, pB, pE );
    p1 = (unsigned char *) getHBScanLine( h-1-y+1, inp, w, bpp );      
    p0 = (unsigned char *) getHBScanLine( h-1-y-0, inp, w, bpp );
    p2 = (unsigned char *) getHBScanLine( h-1-y-1, inp, w, bpp );
    p4 = (unsigned char *) getHBScanLine( h-1-y-2, inp, w, bpp );

    for (x=(pB+2)*3; x<(pE-1)*3; x++)  
      p0[x] = 
      ( 
        ( k[2][1]*p1[x-6] ) + ( k[2][2]*p1[x-3] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+3] ) + ( k[2][5]*p1[x+6] ) +
        ( k[3][1]*p0[x-6] ) + ( k[3][2]*p0[x-3] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+3] ) + ( k[3][5]*p0[x+6] ) +
        ( k[4][1]*p2[x-6] ) + ( k[4][2]*p2[x-3] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+3] ) + ( k[4][5]*p2[x+6] ) +
        ( k[5][1]*p4[x-6] ) + ( k[5][2]*p4[x-3] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+3] ) + ( k[5][5]*p4[x+6] )
      ) / bias1;

    // CENTER X
    for (y=2; y<h-2; y++)
    {
      showProgress( par, y, h, txt );
      if ( testAbort( par ) == 1) break;
      findLineBegEndTopBotI( par, w, y, pI, pB, pE );

      p3 = (unsigned char *) getHBScanLine( h-1-y-2, inp, w, bpp );      
      p1 = (unsigned char *) getHBScanLine( h-1-y-1, inp, w, bpp );      
      p0 = (unsigned char *) getHBScanLine( h-1-y-0, inp, w, bpp );
      p2 = (unsigned char *) getHBScanLine( h-1-y+1, inp, w, bpp );
      p4 = (unsigned char *) getHBScanLine( h-1-y+2, inp, w, bpp );

      // LEFT
      x = 0;
      if (pB <= x)
      for (x=0; x<3; x++)
      p0[x] = 
      ( 
        ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+3] ) + ( k[1][5]*p3[x+6] ) +
        ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+3] ) + ( k[2][5]*p1[x+6] ) +
        ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+3] ) + ( k[3][5]*p0[x+6] ) +
        ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+3] ) + ( k[4][5]*p2[x+6] ) +
        ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+3] ) + ( k[5][5]*p4[x+6] )
      ) / bias2;

      x = 1;
      if (pB <= x)
      for (x=3; x<6; x++)
      p0[x] = 
      ( 
        ( k[1][2]*p3[x-3] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+3] ) + ( k[1][5]*p3[x+6] ) +
        ( k[2][2]*p1[x-3] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+3] ) + ( k[2][5]*p1[x+6] ) +
        ( k[3][2]*p0[x-3] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+3] ) + ( k[3][5]*p0[x+6] ) +
        ( k[4][2]*p2[x-3] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+3] ) + ( k[4][5]*p2[x+6] ) +
        ( k[5][2]*p4[x-3] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+3] ) + ( k[5][5]*p4[x+6] )
      ) / bias1;

      // CENTER
      for (x=(pB+2)*3; x<(pE-1)*3; x++)
        p0[x] = 
        ( 
          ( k[1][1]*p3[x-6] ) + ( k[1][2]*p3[x-3] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+3] ) + ( k[1][5]*p3[x+6] ) +
          ( k[2][1]*p1[x-6] ) + ( k[2][2]*p1[x-3] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+3] ) + ( k[2][5]*p1[x+6] ) +
          ( k[3][1]*p0[x-6] ) + ( k[3][2]*p0[x-3] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+3] ) + ( k[3][5]*p0[x+6] ) +
          ( k[4][1]*p2[x-6] ) + ( k[4][2]*p2[x-3] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+3] ) + ( k[4][5]*p2[x+6] ) +
          ( k[5][1]*p4[x-6] ) + ( k[5][2]*p4[x-3] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+3] ) + ( k[5][5]*p4[x+6] )
        ) / bias;
 
      // RIGHT
      x = w-2;
      if (pE >= x)
      for (x=w*3-6; x<w*3-3; x++)
      p0[x] = 
      ( 
        ( k[1][1]*p3[x-6] ) + ( k[1][2]*p3[x-3] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+3] ) + 
        ( k[2][1]*p1[x-6] ) + ( k[2][2]*p1[x-3] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+3] ) + 
        ( k[3][1]*p0[x-6] ) + ( k[3][2]*p0[x-3] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+3] ) + 
        ( k[4][1]*p2[x-6] ) + ( k[4][2]*p2[x-3] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+3] ) + 
        ( k[5][1]*p4[x-6] ) + ( k[5][2]*p4[x-3] ) + ( k[5][3]*p4[x] ) + ( k[5][4]*p4[x+3] )  
      ) / bias1;

      x = w-1;
      if (pE >= x)
      for (x=w*3-3; x<w*3; x++)
      p0[x] = 
      ( 
        ( k[1][1]*p3[x-6] ) + ( k[1][2]*p3[x-3] ) + ( k[1][3]*p3[x] ) + 
        ( k[2][1]*p1[x-6] ) + ( k[2][2]*p1[x-3] ) + ( k[2][3]*p1[x] ) + 
        ( k[3][1]*p0[x-6] ) + ( k[3][2]*p0[x-3] ) + ( k[3][3]*p0[x] ) + 
        ( k[4][1]*p2[x-6] ) + ( k[4][2]*p2[x-3] ) + ( k[4][3]*p2[x] ) + 
        ( k[5][1]*p4[x-6] ) + ( k[5][2]*p4[x-3] ) + ( k[5][3]*p4[x] )  
      ) / bias2;

    } // for (y=2; y<h-2; y++)


    // BOTTOM
    y = h-2;
    findLineBegEndTopBotI( par, w, y, pI, pB, pE );

    p3 = (unsigned char *) getHBScanLine( h-1-y+2, inp, w, bpp );      
    p1 = (unsigned char *) getHBScanLine( h-1-y+1, inp, w, bpp );      
    p0 = (unsigned char *) getHBScanLine( h-1-y-0, inp, w, bpp );
    p2 = (unsigned char *) getHBScanLine( h-1-y-1, inp, w, bpp );

    for (x=(pB+2)*3; x<(pE-1)*3; x++) 
      p0[x] = 
      ( 
        ( k[1][1]*p3[x-6] ) + ( k[1][2]*p3[x-3] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+3] ) + ( k[1][5]*p3[x+6] ) +
        ( k[2][1]*p1[x-6] ) + ( k[2][2]*p1[x-3] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+3] ) + ( k[2][5]*p1[x+6] ) +
        ( k[3][1]*p0[x-6] ) + ( k[3][2]*p0[x-3] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+3] ) + ( k[3][5]*p0[x+6] ) +
        ( k[4][1]*p2[x-6] ) + ( k[4][2]*p2[x-3] ) + ( k[4][3]*p2[x] ) + ( k[4][4]*p2[x+3] ) + ( k[4][5]*p2[x+6] ) 
      ) / bias1;

    
    y = h-1;
    findLineBegEndTopBotI( par, w, y, pI, pB, pE );

    p3 = (unsigned char *) getHBScanLine( h-1-y+2, inp, w, bpp );      
    p1 = (unsigned char *) getHBScanLine( h-1-y+1, inp, w, bpp );      
    p0 = (unsigned char *) getHBScanLine( h-1-y+0, inp, w, bpp );

    for (x=(pB+2)*3; x<(pE-1)*3; x++) 
      p0[x] = 
      ( 
        ( k[1][1]*p3[x-6] ) + ( k[1][2]*p3[x-3] ) + ( k[1][3]*p3[x] ) + ( k[1][4]*p3[x+3] ) + ( k[1][5]*p3[x+6] ) +
        ( k[2][1]*p1[x-6] ) + ( k[2][2]*p1[x-3] ) + ( k[2][3]*p1[x] ) + ( k[2][4]*p1[x+3] ) + ( k[2][5]*p1[x+6] ) +
        ( k[3][1]*p0[x-6] ) + ( k[3][2]*p0[x-3] ) + ( k[3][3]*p0[x] ) + ( k[3][4]*p0[x+3] ) + ( k[3][5]*p0[x+6] ) 
      ) / bias2;

  } // 24 bpp


}









