/*******************************************************************************

  Filters used by Example Plugin Library (DLL) for DIMIN Viewer n5 (DIV5)

  Programmer: Dima V. Fedorov Levit <dimin@dimin.net> <http://www.dimin.net/>

  History:
    12/05/2000 20:36:48 - First Creation
    03/25/2001 17:11:55 - Conversion to v2.1
    07/07/2004 23:02:00 - Conversion v5
    
*******************************************************************************/

#include <stdio.h>
#include <windows.h>
#include <math.h>

#include <string>
#include <vector>
#include <algorithm>

#include "div5_xtdFilters.h"

#include "DIVPluginDefs.h"
#include "DIVFilterDefs.h"
#include "DIVFilterUtils.h"

#include "colors.h"
//#include "colors.h"
#include "convolve.h"

//------------------------------------------------------------------------------
// Filter 1 rotates provided image by 90 degrees to the left
//------------------------------------------------------------------------------

int __stdcall filterRL( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp ;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  uint x, y;  
  unsigned int w=inbmp.bmHeight, h=inbmp.bmWidth;
  uint bpp = inbmp.bmBitsPixel * inbmp.bmPlanes;
  unsigned char *inbits = (unsigned char *) inbmp.bmBits;


  if ( (bpp == 1) || (bpp == 4) || (bpp == 16) ) return 1;

  //----------------------------------------------
  // now allocate new image in memory
  //----------------------------------------------

  // Allocate memory for the BITMAPINFO structure. (This structure 
  // contains a BITMAPINFOHEADER structure and an array of RGBQUAD 
  // data structures.) 
  BITMAPINFO *pbmi;
  uint pbmi_size = 0;

  if (bpp != 24) 
    pbmi_size = sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * (1<< bpp); 
  else 
    // There is no RGBQUAD array for the 24-bit-per-pixel format. 
    pbmi_size = sizeof(BITMAPINFOHEADER);

  pbmi = (PBITMAPINFO) LocalAlloc(LPTR, pbmi_size ); 
  memset(pbmi, 0, pbmi_size );


  // Initialize the fields in the BITMAPINFO structure. 
  pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER); 
  pbmi->bmiHeader.biWidth = w; 
  pbmi->bmiHeader.biHeight = h; 
  pbmi->bmiHeader.biPlanes = 1; 
  pbmi->bmiHeader.biBitCount = bpp; 

  if (bpp < 24) 
    pbmi->bmiHeader.biClrUsed = (1<<bpp);
  else
    pbmi->bmiHeader.biClrUsed = 0;

  // If the bitmap is not compressed, set the BI_RGB flag. 
  pbmi->bmiHeader.biCompression = BI_RGB; 

  // Compute the number of bytes in the array of color 
  // indices and store the result in biSizeImage. 
  // For Windows NT, the width must be DWORD aligned unless 
  // the bitmap is RLE compressed. This example shows this. 
  // For Windows 95/98/Me, the width must be WORD aligned unless the 
  // bitmap is RLE compressed.
  pbmi->bmiHeader.biSizeImage = ((w * bpp +31) & ~31) /8 * h; 
  // Set biClrImportant to 0, indicating that all of the 
  // device colors are important. 
  pbmi->bmiHeader.biClrImportant = 0; 

  //----------------------------------------------
  // create palette if needed
  //----------------------------------------------

  if (bpp != 24)
  {
    int i, pals;
    PALETTEENTRY pal[256];

    pals = GetPaletteEntries( par->hpal, 0, 256, pal );

    if (pals > 0)
      for (i=0; i<256; i++)
      {
        pbmi->bmiColors[i].rgbRed   = pal[i].peRed;
        pbmi->bmiColors[i].rgbGreen = pal[i].peGreen;
        pbmi->bmiColors[i].rgbBlue  = pal[i].peBlue;
      }
    else
      for (i=0; i<256; i++)
      {
        pbmi->bmiColors[i].rgbRed   = (unsigned char) i;
        pbmi->bmiColors[i].rgbGreen = (unsigned char) i;
        pbmi->bmiColors[i].rgbBlue  = (unsigned char) i;
      }

  } // if paletted
  
  unsigned char *p;
  par->hbmp = CreateDIBSection(NULL, pbmi, DIB_RGB_COLORS, (void**)&p, NULL, 0);
  GdiFlush();
  if (p == NULL) return 1;


  unsigned char *inP  = inbits;
  unsigned char *outP = p;
  
  // now copy image rotating it
  unsigned int inLineSize = inbmp.bmWidthBytes;
  unsigned int outLineSize = ((w * bpp +31) & ~31) /8;

  try
  {

  if (bpp == 24)
  for (y=0; y<w; y++) // we go by old image height
  {
    showProgress( par, y, w, "Left rotation" );
    if ( testAbort( par ) == 1) break;

    outP = p+(((w-1)-y)*3);
    unsigned char *ip = inP;
    for (x=0; x<h; x++) // we go by old image width
    {
      outP[0] = *ip; ip++;
      outP[1] = *ip; ip++;
      outP[2] = *ip; ip++;
      outP += outLineSize;
    }
    inP += inLineSize;
  }

  /*
  if (bpp == 16)
  for (y=0; y<w; y++) // we go by old image height
  {
    showProgress( par, y, w, "Left rotation" );
    if ( testAbort( par ) == 1) break;

    outP = p+(((w-1)-y)*2);
    unsigned char *ip = inP;
    for (x=0; x<h; x++) // we go by old image width
    {
      outP[0] = *ip; ip++;
      outP[1] = *ip; ip++;
      outP[2] = *ip; ip++;
      outP += outLineSize;
    }
    inP += inLineSize;
  }
  */

  if (bpp == 8)
  for (y=0; y<w; y++) // we go by old image height
  {
    showProgress( par, y, w, "Left rotation" );
    if ( testAbort( par ) == 1) break;

    outP = p+((w-1)-y);
    unsigned char *ip = inP;
    for (x=0; x<h; x++) // we go by old image width
    {
      outP[0] = *ip; ip++;
      outP += outLineSize;
    }
    inP += inLineSize;
  }

  /*
  if (bpp == 4)
  {
    w = floor(w/2);
    for (y=0; y<w; y++) // we go by old image height
    {
      showProgress( par, y, w, "Left rotation" );
      if ( testAbort( par ) == 1) break;

      outP = p+((w-1)-y);
      unsigned char *ip = inP;
      for (x=0; x<h; x++) // we go by old image width
      {
        outP[0] = *ip; ip++;
        outP += outLineSize;
      }
      inP += inLineSize;
    }
  } // 4 bpp

  if (bpp == 1)
  {
    w = ceil(w/8);
    h = ceil(h/8);
    for (y=0; y<w; y++) // we go by old image height
    {
      showProgress( par, y, w, "Left rotation" );
      if ( testAbort( par ) == 1) break;

      outP = p+((w-1)-y);
      unsigned char *ip = inP;
      for (x=0; x<h; x++) // we go by old image width
      {
        outP[0] = *ip; ip++;
        outP += outLineSize;
      }
      inP += inLineSize;
    }
  } // 1 bpp
*/

  if (inhbmp != NULL) DeleteObject( inhbmp );

  GdiFlush();

  }
  catch ( ... )
  {
    return 1;
  }

  return 0;

}

//------------------------------------------------------------------------------
// Filter 2 rotates provided image by 90 degrees to the right
//------------------------------------------------------------------------------

int __stdcall filterRR( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp ;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  uint x, y;  
  unsigned int w=inbmp.bmHeight, h=inbmp.bmWidth;
  uint bpp = inbmp.bmBitsPixel * inbmp.bmPlanes;
  unsigned char *inbits = (unsigned char *) inbmp.bmBits;

  if ( (bpp == 1) || (bpp == 4) || (bpp == 16) ) return 1;

  //----------------------------------------------
  // now allocate new image in memory
  //----------------------------------------------

  // Allocate memory for the BITMAPINFO structure. (This structure 
  // contains a BITMAPINFOHEADER structure and an array of RGBQUAD 
  // data structures.) 
  BITMAPINFO *pbmi;
  uint pbmi_size = 0;

  if (bpp != 24) 
    pbmi_size = sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * (1<< bpp); 
  else 
    // There is no RGBQUAD array for the 24-bit-per-pixel format. 
    pbmi_size = sizeof(BITMAPINFOHEADER);

  pbmi = (PBITMAPINFO) LocalAlloc(LPTR, pbmi_size ); 
  memset(pbmi, 0, pbmi_size );


  // Initialize the fields in the BITMAPINFO structure. 
  pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER); 
  pbmi->bmiHeader.biWidth = w; 
  pbmi->bmiHeader.biHeight = h; 
  pbmi->bmiHeader.biPlanes = 1; 
  pbmi->bmiHeader.biBitCount = bpp; 

  if (bpp < 24) 
    pbmi->bmiHeader.biClrUsed = (1<<bpp);
  else
    pbmi->bmiHeader.biClrUsed = 0;

  // If the bitmap is not compressed, set the BI_RGB flag. 
  pbmi->bmiHeader.biCompression = BI_RGB; 

  // Compute the number of bytes in the array of color 
  // indices and store the result in biSizeImage. 
  // For Windows NT, the width must be DWORD aligned unless 
  // the bitmap is RLE compressed. This example shows this. 
  // For Windows 95/98/Me, the width must be WORD aligned unless the 
  // bitmap is RLE compressed.
  pbmi->bmiHeader.biSizeImage = ((w * bpp +31) & ~31) /8 * h; 
  // Set biClrImportant to 0, indicating that all of the 
  // device colors are important. 
  pbmi->bmiHeader.biClrImportant = 0; 

  //----------------------------------------------
  // create palette if needed
  //----------------------------------------------

  if (bpp != 24)
  {
    int i, pals;
    PALETTEENTRY pal[256];

    pals = GetPaletteEntries( par->hpal, 0, 256, pal );

    if (pals > 0)
      for (i=0; i<256; i++)
      {
        pbmi->bmiColors[i].rgbRed   = pal[i].peRed;
        pbmi->bmiColors[i].rgbGreen = pal[i].peGreen;
        pbmi->bmiColors[i].rgbBlue  = pal[i].peBlue;
      }
    else
      for (i=0; i<256; i++)
      {
        pbmi->bmiColors[i].rgbRed   = (unsigned char) i;
        pbmi->bmiColors[i].rgbGreen = (unsigned char) i;
        pbmi->bmiColors[i].rgbBlue  = (unsigned char) i;
      }

  } // if paletted
  
  unsigned char *p;
  par->hbmp = CreateDIBSection(NULL, pbmi, DIB_RGB_COLORS, (void**)&p, NULL, 0);
  GdiFlush();
  if (p == NULL) return 1;


  unsigned char *inP  = inbits;
  unsigned char *outP = p;
  
  // now copy image rotating it
  unsigned int inLineSize = inbmp.bmWidthBytes;
  unsigned int outLineSize = ((w * bpp +31) & ~31) /8;

  try
  {

  if (bpp == 24)
  for (y=0; y<w; y++) // we go by old image height
  {
    showProgress( par, y, w, "Right rotation" );
    if ( testAbort( par ) == 1) break;

    outP = p+y*3;
    unsigned char *ip = inP+(h-1)*3;
    for (x=0; x<h; x++) // we go by old image width
    {
      outP[0] = ip[0];
      outP[1] = ip[1];
      outP[2] = ip[2];
      ip-=3;
      outP += outLineSize;
    }
    inP += inLineSize;
  }

  /*
  if (bpp == 16)
  for (y=0; y<w; y++) // we go by old image height
  {
    showProgress( par, y, w, "Right rotation" );
    if ( testAbort( par ) == 1) break;

    outP = p+y*2;
    unsigned char *ip = inP;
    for (x=0; x<h; x++) // we go by old image width
    {
      outP[0] = *ip; ip++;
      outP[1] = *ip; ip++;
      outP[2] = *ip; ip++;
      outP += outLineSize;
    }
    inP += inLineSize;
  }
  */

  if (bpp == 8)
  for (y=0; y<w; y++) // we go by old image height
  {
    showProgress( par, y, w, "Right rotation" );
    if ( testAbort( par ) == 1) break;

    outP = p+y;
    unsigned char *ip = inP+(h-1);
    for (x=0; x<h; x++) // we go by old image width
    {
      outP[0] = *ip; ip--;
      outP += outLineSize;
    }
    inP += inLineSize;
  }


  if (inhbmp != NULL) DeleteObject( inhbmp );

  GdiFlush();

  }
  catch ( ... )
  {
    return 1;
  }

  return 0;

}

//------------------------------------------------------------------------------
// Filter 3 find edges
//------------------------------------------------------------------------------

int __stdcall filterFindEdg( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  uint x, y;  
  unsigned int h=inbmp.bmHeight, w=inbmp.bmWidth;
  uint bpp = inbmp.bmBitsPixel * inbmp.bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp.bmBits;

  try
  {
    uint pB=0, pE=w-1;
    uint pI=0;
  
    for (y=h-1; y>0; y--) // we go by old image height
    {
      showProgress( par, h-y-1, h, "Find Edges" );
      if ( testAbort( par ) == 1) break;
      findLineBegEndTopBot( par, w-1, h-y-1, pI, pB, pE);
      unsigned char *p  = getHBScanLine( y, inp, w, bpp );
      unsigned char *p2 = getHBScanLine( y-1, inp, w, bpp );

      if (bpp == 8)
      for (x=pB; x<=pE; x++)
      {
        p[x] = ( abs(p[x] - p[x+1]) + abs(p[x] - p2[x]) ) / 2;
        if (y == 1) p2[x] = 0;
      }

      if (bpp == 24)
      for (x=pB*3; x<=pE*3; x+=3)
      {
        p[x+0] = ( abs(p[x+0] - p[x+3]) + abs(p[x+0] - p2[x+0]) ) / 2;
        p[x+1] = ( abs(p[x+1] - p[x+4]) + abs(p[x+1] - p2[x+1]) ) / 2;
        p[x+2] = ( abs(p[x+2] - p[x+5]) + abs(p[x+2] - p2[x+2]) ) / 2;
        if (y==1) { p2[x+0] = 0; p2[x+1] = 0; p2[x+2] = 0; }
      }

    }

  }
  catch ( ... )
  {
    return 1;
  }

  return 0;

}


//------------------------------------------------------------------------------
// Filter 4 median
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// median filter
//------------------------------------------------------------------------------

void div_median3x3_fine( TDIVParameters *par, BITMAP *inbmp, int amount )
{
  std::vector<int> v(9);  
  unsigned char /* *p0, *p1, *p2, */ *pr;
  long x,y;

  unsigned int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;

  DIM_UCHAR *tp;
  HPALETTE hpal;
  HBITMAP temp = dimCreateHBitmap( w, h, bpp, 0, NULL, &tp, hpal );
  
  int i;
  for (i=0; i<amount; i++)
  {
    uint pB=0, pE=w-1;
    uint pI=0;
    memcpy( tp, inp, getHBLineSizeInBytes(w, bpp) * h );

    for (y=h-2; y>1; y--) // we go by old image height
    {
      showProgress( par, h-y-1, h, "Median" );
      if ( testAbort( par ) == 1) break;
      findLineBegEndTopBot( par, w-1, h-y-1, pI, pB, pE);

      unsigned char *pixB = getHBScanLine( y-1, tp, w, bpp );
      unsigned char *pixC = getHBScanLine( y,   tp, w, bpp );
      unsigned char *pixT = getHBScanLine( y+1, tp, w, bpp );
      pr = (unsigned char *)getHBScanLine( y,   inp, w, bpp );

      if (bpp == 8)
      for (x=pB+1; x<pE; x++)
      {
        v.at(0) = pixC[x-1]; v.at(1) = pixC[x]; v.at(2) = pixC[x+1];
        v.at(3) = pixB[x-1]; v.at(4) = pixB[x]; v.at(5) = pixB[x+1];
        v.at(6) = pixT[x-1]; v.at(7) = pixT[x]; v.at(8) = pixT[x+1];
        std::sort(v.begin(), v.end());
        pr[x] = v.at(4);
      }

      if (bpp == 24)
      for (x=(pB+1)*3; x<pE*3; x+=3)
      {
        v.at(0) = pixC[x-3]; v.at(1) = pixC[x]; v.at(2) = pixC[x+3];
        v.at(3) = pixB[x-3]; v.at(4) = pixB[x]; v.at(5) = pixB[x+3];
        v.at(6) = pixT[x-3]; v.at(7) = pixT[x]; v.at(8) = pixT[x+3];
        std::sort(v.begin(), v.end());
        pr[x] = v.at(4);

        v.at(0) = pixC[x-2]; v.at(1) = pixC[x+1]; v.at(2) = pixC[x+4];
        v.at(3) = pixB[x-2]; v.at(4) = pixB[x+1]; v.at(5) = pixB[x+4];
        v.at(6) = pixT[x-2]; v.at(7) = pixT[x+1]; v.at(8) = pixT[x+4];
        std::sort(v.begin(), v.end());
        pr[x+1] = v.at(4);

        v.at(0) = pixC[x-1]; v.at(1) = pixC[x+2]; v.at(2) = pixC[x+5];
        v.at(3) = pixB[x-1]; v.at(4) = pixB[x+2]; v.at(5) = pixB[x+5];
        v.at(6) = pixT[x-1]; v.at(7) = pixT[x+2]; v.at(8) = pixT[x+5];
        std::sort(v.begin(), v.end());
        pr[x+2] = v.at(4);
      }
    }


  } // for amount

  DeleteObject( temp );
}

void div_median3x3(TDIVParameters *par, BITMAP *inbmp)
{
  std::vector<int> v(9);

  uint x, y;  
  unsigned int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;


  uint pB=0, pE=w-1;
  uint pI=0;

  for (y=h-2; y>1; y--) // we go by old image height
  {
    showProgress( par, h-y-1, h, "Median" );
    if ( testAbort( par ) == 1) break;
    findLineBegEndTopBot( par, w-1, h-y-1, pI, pB, pE);
    unsigned char *pixB = getHBScanLine( y-1, inp, w, bpp );
    unsigned char *pixC = getHBScanLine( y,   inp, w, bpp );
    unsigned char *pixT = getHBScanLine( y+1, inp, w, bpp );

    if (bpp == 8)
    for (x=pB+1; x<pE; x++)
    {
      v.at(0) = pixC[x-1]; v.at(1) = pixC[x]; v.at(2) = pixC[x+1];
      v.at(3) = pixB[x-1]; v.at(4) = pixB[x]; v.at(5) = pixB[x+1];
      v.at(6) = pixT[x-1]; v.at(7) = pixT[x]; v.at(8) = pixT[x+1];
      std::sort(v.begin(), v.end());
      pixC[x] = v.at(4);
    }

    if (bpp == 24)
    for (x=(pB+1)*3; x<pE*3; x+=3)
    {
      v.at(0) = pixC[x-3]; v.at(1) = pixC[x]; v.at(2) = pixC[x+3];
      v.at(3) = pixB[x-3]; v.at(4) = pixB[x]; v.at(5) = pixB[x+3];
      v.at(6) = pixT[x-3]; v.at(7) = pixT[x]; v.at(8) = pixT[x+3];
      std::sort(v.begin(), v.end());
      pixC[x] = v.at(4);

      v.at(0) = pixC[x-2]; v.at(1) = pixC[x+1]; v.at(2) = pixC[x+4];
      v.at(3) = pixB[x-2]; v.at(4) = pixB[x+1]; v.at(5) = pixB[x+4];
      v.at(6) = pixT[x-2]; v.at(7) = pixT[x+1]; v.at(8) = pixT[x+4];
      std::sort(v.begin(), v.end());
      pixC[x+1] = v.at(4);

      v.at(0) = pixC[x-1]; v.at(1) = pixC[x+2]; v.at(2) = pixC[x+5];
      v.at(3) = pixB[x-1]; v.at(4) = pixB[x+2]; v.at(5) = pixB[x+5];
      v.at(6) = pixT[x-1]; v.at(7) = pixT[x+2]; v.at(8) = pixT[x+5];
      std::sort(v.begin(), v.end());
      pixC[x+2] = v.at(4);
    }
  }

}

int __stdcall filterMedianStrong( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    unsigned int amount=1;
    getUIntArg ( par, 0, amount );
    
    for (int i=0; i<amount; i++)
      div_median3x3(par, &inbmp);
  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}

int __stdcall filterMedian( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    unsigned int amount=1;
    getUIntArg ( par, 0, amount );
    div_median3x3_fine(par, &inbmp, amount);
  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}

//******************************************************************************
// Filter 5 blur 3x3
//******************************************************************************

int __stdcall filterBlur3x3( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    unsigned int amount=1, i;
    int m[16], *k[4];
    k[0] = &m[0]; k[1] = &m[4]; k[2] = &m[8]; k[3] = &m[12];

    /*
    k[1][1] = 1; k[1][2] = 2; k[1][3] = 1;
    k[2][1] = 2; k[2][2] = 4; k[2][3] = 2;
    k[3][1] = 1; k[3][2] = 2; k[3][3] = 1;

    k[1][1] = 1; k[1][2] = 1; k[1][3] = 1;
    k[2][1] = 1; k[2][2] = 1; k[2][3] = 1;
    k[3][1] = 1; k[3][2] = 1; k[3][3] = 1;
    */

    k[1][1] = 1; k[1][2] = 1; k[1][3] = 1;
    k[2][1] = 1; k[2][2] = 1; k[2][3] = 1;
    k[3][1] = 1; k[3][2] = 1; k[3][3] = 1;

    getUIntArg ( par, 0, amount );
      
    for (i=0; i<amount; i++)
      applyKernel33( par, &inbmp, (int **) k, 9, 6, "Smooth" );
  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}

//******************************************************************************
// Filter 6 blur gauss 5x5
//******************************************************************************

int __stdcall filterBlurGauss( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    unsigned int amount=1, i, j;
    int m[36], *k[6];
    k[0] = &m[0]; k[1] = &m[6]; k[2] = &m[12]; k[3] = &m[18]; k[4] = &m[24]; k[5] = &m[30];

    /*
    k[1][1] = 0; k[1][2] = 0;    k[1][3] = 0;    k[1][4] = 0;    k[1][5] = 0;
    k[2][1] = 0; k[2][2] = 625;  k[2][3] = 1250; k[2][4] = 625;  k[2][5] = 0;
    k[3][1] = 0; k[3][2] = 1250; k[3][3] = 2500; k[3][4] = 1250; k[3][5] = 0;
    k[4][1] = 0; k[4][2] = 625;  k[4][3] = 1250; k[4][4] = 625;  k[4][5] = 0;
    k[5][1] = 0; k[5][2] = 0;    k[5][3] = 0;    k[5][4] = 0;    k[5][5] = 0;

    k[1][1] = 25;  k[1][2] = 125;  k[1][3] = 200;  k[1][4] = 125;  k[1][5] = 25;
    k[2][1] = 125; k[2][2] = 625;  k[2][3] = 1000; k[2][4] = 625;  k[2][5] = 125;
    k[3][1] = 200; k[3][2] = 1000; k[3][3] = 1600; k[3][4] = 1000; k[3][5] = 200;
    k[4][1] = 125; k[4][2] = 625;  k[4][3] = 1000; k[4][4] = 625;  k[4][5] = 125;
    k[5][1] = 25;  k[5][2] = 125;  k[5][3] = 200;  k[5][4] = 125;  k[5][5] = 25;


    k[1][1] = 2; k[1][2] = 4;  k[1][3] = 5;  k[1][4] = 4;  k[1][5] = 2;
    k[2][1] = 4; k[2][2] = 9;  k[2][3] = 12; k[2][4] = 9;  k[2][5] = 4;
    k[3][1] = 5; k[3][2] = 12; k[3][3] = 15; k[3][4] = 12; k[3][5] = 5;
    k[4][1] = 4; k[4][2] = 9;  k[4][3] = 12; k[4][4] = 9;  k[4][5] = 4;
    k[5][1] = 2; k[5][2] = 4;  k[5][3] = 5;  k[5][4] = 4;  k[5][5] = 2;
    */

    k[1][1] = 25;  k[1][2] = 125;  k[1][3] = 200;  k[1][4] = 125;  k[1][5] = 25;
    k[2][1] = 125; k[2][2] = 625;  k[2][3] = 1000; k[2][4] = 625;  k[2][5] = 125;
    k[3][1] = 200; k[3][2] = 1000; k[3][3] = 1600; k[3][4] = 1000; k[3][5] = 200;
    k[4][1] = 125; k[4][2] = 625;  k[4][3] = 1000; k[4][4] = 625;  k[4][5] = 125;
    k[5][1] = 25;  k[5][2] = 125;  k[5][3] = 200;  k[5][4] = 125;  k[5][5] = 25;

    int bias=0, bias1=0, bias2=0;
    for (i=1; i<6; i++)
      for (j=1; j<6; j++)
        bias += k[i][j];

    for (i=2; i<6; i++)
      for (j=1; j<6; j++)
        bias1 += k[i][j];

    for (i=3; i<6; i++)
      for (j=1; j<6; j++)
        bias2 += k[i][j];

    getUIntArg ( par, 0, amount );
      
    for (i=0; i<amount; i++)
      applyKernel55( par, &inbmp, (int **) k, bias, bias1, bias2, "Gaussian Blur" );

  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}


//******************************************************************************
// Filter 7 RGB -> HSL
//******************************************************************************

void div_rgb_hsl( TDIVParameters *par, BITMAP *inbmp )
{
  unsigned char *p;
  uint x, y;  
  unsigned int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;

  if (bpp != 24) return;

  uint pB=0, pE=w-1;
  uint pI=0;

  for (y=0; y<h; y++)
  {
    showProgress( par, y, h, "RGB->HSL" );
    if ( testAbort( par ) == 1) break;
    findLineBegEndTopBot( par, w, y, pI, pB, pE );

    p = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );

    for (x=pB*3; x<=pE*3; x+=3)
      RGBtoHSL( p[x+2], p[x+1], p[x], p[x+2], p[x+1], p[x] );
  } 
}

int __stdcall filterRGBtoHSL( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    div_rgb_hsl( par, &inbmp );
  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}


//******************************************************************************
// Filter 8 HSL -> RGB
//******************************************************************************

void div_hsl_rgb( TDIVParameters *par, BITMAP *inbmp )
{
  unsigned char *p;
  uint x, y;  
  unsigned int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;

  if (bpp != 24) return;

  uint pB=0, pE=w-1;
  uint pI=0;

  for (y=0; y<h; y++)
  {
    showProgress( par, y, h, "RGB->HSL" );
    if ( testAbort( par ) == 1) break;
    findLineBegEndTopBot( par, w, y, pI, pB, pE );

    p = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );

    for (x=pB*3; x<=pE*3; x+=3)
      HSLtoRGB( p[x+2], p[x+1], p[x], p[x+2], p[x+1], p[x] );
  } 
}

int __stdcall filterHSLtoRGB( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    div_hsl_rgb( par, &inbmp );
  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}


//******************************************************************************
// Filter 9 RGB -> YUV
//******************************************************************************

void div_rgb_yuv( TDIVParameters *par, BITMAP *inbmp )
{
  unsigned char *p;
  uint x, y;  
  unsigned int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;

  if (bpp != 24) return;

  uint pB=0, pE=w-1;
  uint pI=0;

  for (y=0; y<h; y++)
  {
    showProgress( par, y, h, "RGB->YUV" );
    if ( testAbort( par ) == 1) break;
    findLineBegEndTopBot( par, w, y, pI, pB, pE );

    p = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );

    for (x=pB*3; x<=pE*3; x+=3)
      RGBtoYUV( p[x+2], p[x+1], p[x], p[x+2], p[x+1], p[x] );
  } 
}

int __stdcall filterRGBtoYUV( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    div_rgb_yuv( par, &inbmp );
  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}


//******************************************************************************
// Filter 10 YUV -> RGB
//******************************************************************************

void div_yuv_rgb( TDIVParameters *par, BITMAP *inbmp )
{
  unsigned char *p;
  uint x, y;  
  unsigned int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;

  if (bpp != 24) return;

  uint pB=0, pE=w-1;
  uint pI=0;

  for (y=0; y<h; y++)
  {
    showProgress( par, y, h, "RGB->YUV" );
    if ( testAbort( par ) == 1) break;
    findLineBegEndTopBot( par, w, y, pI, pB, pE );

    p = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );

    for (x=pB*3; x<=pE*3; x+=3)
      YUVtoRGB( p[x+2], p[x+1], p[x], p[x+2], p[x+1], p[x] );
  } 
}

int __stdcall filterYUVtoRGB( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    div_yuv_rgb( par, &inbmp );
  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}


//******************************************************************************
// Filter 11 RGB -> CIE La*b*
//******************************************************************************

void div_rgb_lab( TDIVParameters *par, BITMAP *inbmp )
{
  unsigned char *p;
  uint x, y;  
  unsigned int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;

  if (bpp != 24) return;

  uint pB=0, pE=w-1;
  uint pI=0;

  for (y=0; y<h; y++)
  {
    showProgress( par, y, h, "RGB->LAB" );
    if ( testAbort( par ) == 1) break;
    findLineBegEndTopBot( par, w, y, pI, pB, pE );

    p = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );

    for (x=pB*3; x<=pE*3; x+=3)
      RGBtoLAB( p[x+2], p[x+1], p[x], p[x+2], p[x+1], p[x] );
  } 

}

int __stdcall filterRGBtoLAB( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    div_rgb_lab( par, &inbmp );
  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}

//******************************************************************************
// Filter 12 CIE La*b* -> RGB
//******************************************************************************

void div_lab_rgb( TDIVParameters *par, BITMAP *inbmp )
{
  unsigned char *p;
  uint x, y;  
  unsigned int h=inbmp->bmHeight, w=inbmp->bmWidth;
  uint bpp = inbmp->bmBitsPixel * inbmp->bmPlanes;
  unsigned char *inp = (unsigned char *) inbmp->bmBits;

  if (bpp != 24) return;

  uint pB=0, pE=w-1;
  uint pI=0;

  for (y=0; y<h; y++)
  {
    showProgress( par, y, h, "LAB->RGB" );
    if ( testAbort( par ) == 1) break;
    findLineBegEndTopBot( par, w, y, pI, pB, pE );

    p = (unsigned char *) getHBScanLine( h-y-1, inp, w, bpp );

    for (x=pB*3; x<=pE*3; x+=3)
      LABtoRGB( p[x+2], p[x+1], p[x], p[x+2], p[x+1], p[x] );
  } 
}

int __stdcall filterLABtoRGB( TDIVParameters *par )
{
  if ( par->ver < sizeof(TDIVParameters) ) return 1;
  if (par->hbmp == 0) return 1;

  //----------------------------------------------
  // first obtain input image parameters
  //----------------------------------------------
  HBITMAP inhbmp = par->hbmp;
  BITMAP inbmp;
  GetObject(par->hbmp, sizeof(BITMAP), (LPSTR)&inbmp);

  try
  {
    div_lab_rgb( par, &inbmp );
  }
  catch ( ... )
  {
    return 1;
  }

  return 0;
}









